/**
 * Copyright (c) 2004 Logos GmbH Rostock (www.logos-ib.de).
 * All rights reserved.
 *
 * THIS SOFTWARE IS  PROVIDED  "AS IS"  AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF  MERCHANTABILITY  AND  FITNESS  FOR  A PARTICULAR  PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL  LOGOS GMBH  OR ITS CONTRIBUTORS BE
 * LIABLE FOR ANY  DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
 * OR CONSEQUENTIAL DAMAGES  (INCL., BUT NOT LIMITED TO, PROCUREMENT
 * OF SUBSTITUTE GOODS OR SERVICES;  LOSS OF USE, DATA, OR  PROFITS;
 * OR BUSINESS INTERRUPTION)  HOWEVER  CAUSED  AND  ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,  OR TORT (INCL.
 * NEGLIGENCE OR OTHERWISE)  ARISING  IN  ANY  WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
package de.inovat.kex.bast.common.objectwrapper;

/*
 * Revision 1.3  2007/11/26 12:10:19  Trier
 * <Kein Kommentar eingegeben>
 */

import java.util.ArrayList;
import java.util.List;
import java.util.WeakHashMap;

import de.bsvrz.dav.daf.main.Data;
import de.bsvrz.dav.daf.main.config.AttributeGroup;
import de.bsvrz.dav.daf.main.config.ConfigurationObject;
import de.bsvrz.dav.daf.main.config.ConfigurationObjectType;
import de.bsvrz.dav.daf.main.config.DataModel;
import de.bsvrz.dav.daf.main.config.ObjectSet;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.inovat.kex.bast.common.*;

/**
 * kapselt den DatenModellTyp eak
 * 
 * @author LOGOS GmbH Rostock (www.logos-ib.de)
 * @author Ralf Trier
 * @version $Id: EAK.java 1292 2009-03-20 13:12:46Z Trier $
 */
public class EAK extends ConfigurationObjectWrapper {

  private static WeakHashMap<String,EAK> __Cache = new WeakHashMap<String,EAK>();

  private static ConfigurationObjectType __ConfigurationObjectType;
  private static ConfigurationObjectType __ConfigurationObjectTypeOfDE;
  private static AttributeGroup __AtgEAK;

  private String _Bezeichnung;
  private int _OSI2Adresse;
  private String _UmsetzungsModul;

  private ArrayList<DEKanal> _DEList;
  private DEKanal _DEKanal;
  private SteuerModul _SteuerModul;

  /**
   * setzt das SteuerModul, an den der EAK angeschlossen ist.
   *
   * @param steuerModul SteuerModul, an den der EAK angeschlossen ist.
   */
  public void setSteuerModul(SteuerModul steuerModul) {
    _SteuerModul = steuerModul;
  }

  /**
   * gibt das SteuerModul zur&uuml;ck, an den der SteuerModul angeschlossen ist, und dieses
   * gesetzt wurde.
   *
   * @return SteuerModul, an den der EAK angeschlossen ist.
   */
  public SteuerModul getSteuerModul() {
    return _SteuerModul;
  }

  /**
   * gibt die Bezeichnung zur&uuml;ck.
   *
   * @return String
   */
  public String getBezeichnung() {
    return _Bezeichnung;
  }

  /**
   * gibt die EAK-Port-Nummer auf dem Lokalbus zur&uuml;ck.
   *
   * @return die OSI2Adresse
   */
  public int getOSI2Adresse() {
    return _OSI2Adresse;
  }

  /**
   * gibt  das fr die Umsetzung des Protokolls (Lesen und Schreiben) verwendete
   * Modul an diesem Knoten zur&uuml;ck.
   *
   * @return Klassennamen des UmsetzungsModuls.
   */
  public String getUmsetzungsModul() {
    return _UmsetzungsModul;
  }

  /**
   * gibt die DE-Kan&auml;le zur&uuml;ck oder null, wenn keine angeschlossen sind.
   *
   * @return DE-Kan&auml;le.
   */
  public DEKanal[] getDEKanals() {
    if ((_DEList!=null) && (_DEList.size()>0)) {
      return _DEList.toArray(new DEKanal[_DEList.size()]);
    }
    else {
      return null;
    }
  }


  /**
   * erzeugt das WrapperObject f&uuml;r das Object mit der angegebenen Pid
   *
   * @see ConfigurationObjectWrapper#ConfigurationObjectWrapper(String pid)
   *
   * @param pid String
   * @throws DataModellMismatchException
   * @throws DataModellObjectNotFoundException
   */
  private EAK(String pid) throws DataModellMismatchException, DataModellObjectNotFoundException {
    super(pid);
  }

  /**
   * erzeugt das WrapperObject f&uuml;r das angegebene Object.
   *
   * @see de.inovat.kex.bast.common.objectwrapper.ConfigurationObjectWrapper#ConfigurationObjectWrapper(ConfigurationObject confObj)
   *
   * @param confObj zu wrappenes ConfigurationObject, das ein EAK repr&auml;sentieren muss.
   * @throws DataModellMismatchException
   */
  private EAK(ConfigurationObject confObj) throws DataModellMismatchException {
    super(confObj);
  }

  protected void initData(ConfigurationObject confObj) throws DataModellMismatchException {
    Data oData = confObj.getConfigurationData(__AtgEAK);
    if (oData != null) {
      _Bezeichnung = oData.getTextValue("Bezeichnung").getText();
      _OSI2Adresse = oData.getUnscaledValue("OSI2Adresse").intValue();
      _UmsetzungsModul = oData.getTextValue("UmsetzungsModul").getText();
    }
    ObjectSet os = confObj.getObjectSet("De");
    if (os != null) {
      List<SystemObject> list = os.getElements();
      if ((list!=null) && (list.size()>0)) {
        _DEList = new ArrayList<DEKanal>(list.size());
        for (int i = 0; i < list.size(); i++) {
          ConfigurationObject deObj = (ConfigurationObject)list.get(i);
          if (deObj.isOfType(__ConfigurationObjectTypeOfDE)) {
            DEKanal deKanal = DEKanal.getInstance(deObj);
            deKanal.setEAK(this);
            _DEList.add(deKanal);
          }
        }
      }
    }
    __Cache.put(this.getPID(), this);
  }

  ConfigurationObjectType getType() {
    return __ConfigurationObjectType;
  }

  /**
   * gibt den DEKanal zur Steurung des EAK's zur&uuml;ck.
   *
   * @return DEKanal oder null, wenn der EAK ein virtueller ist (EAK254 OsiAddr=0).
   */
  public DEKanal getSteuerDEKanal() {
    return _DEKanal;
  }

  /**
   * setzt den DEKanal zur Steurung des EAK's (nur, wenn der EAK nicht virtuelle
   * ist).
   *
   * @param deKanal DEKanal &uuml;ber den auf den EAK zugegriffen werden kann.
   */
  public void setDEKanal(DEKanal deKanal) {
    _DEKanal = deKanal;
  }


  public DEKanal findDEKanal(int osiAddr) {

    for (int i=0; i<_DEList.size(); i++) {
      DEKanal deKanal = _DEList.get(i);
      if ( deKanal.getDEKanal() == osiAddr ) {
       return  deKanal;
      }
    }
    return null;
  }



  public String getTreeString() {
    String s = getBezeichnung();
    int index = s.indexOf("(", 2);
    if (index>0) {
      return s.substring(0, index);
    }
    return s;
  }

  /**
   * gibt den EAK mit der &uuml;bergebenen Pid zur&uuml;ck oder null,
   * wenn es kein EAK mit der Pid gibt.
   *
   * @param pid String
   * @throws DataModellMismatchException
   * @return EAK
   */
  public static EAK getInstance(String pid) throws DataModellMismatchException {
    EAK eak = __Cache.get(pid);
    if (eak==null) {
      Object object = NetzClientDavConnection.getInstance().getDataModel().getObject(pid);
      if ((object!=null) && (object instanceof ConfigurationObject)) {
        eak = getInstanceIntern((ConfigurationObject)object);
      }
    }
    return eak;
  }

  /**
   * <p>Liefert das Wrapper-Objekt zum angegebenen Konfigurationsobjekt.</p>
   *
   * @param confObj ConfigurationObject
   * @return EAK
   * @throws DataModellMismatchException
   */
  static public EAK getInstance(ConfigurationObject confObj) throws DataModellMismatchException {
    EAK eak = __Cache.get(confObj.getPid());
    if (eak == null) {
      eak = getInstanceIntern(confObj);
    }
    return eak;
  }

  static synchronized private EAK getInstanceIntern(ConfigurationObject confObj) throws DataModellMismatchException {
    EAK eak = __Cache.get(confObj.getPid());
    if (eak == null) {
      if (__ConfigurationObjectType == null) {
        EAK.init();
      }
      eak = new EAK(confObj);
    }
    return eak;
  }


  private static synchronized void init() {
    if (__ConfigurationObjectType==null) {
      DataModel model = NetzClientDavConnection.getInstance().getDataModel();
      __ConfigurationObjectType = (ConfigurationObjectType)model.getType("typ.eak");

      __AtgEAK = model.getAttributeGroup("atg.eak");
      __ConfigurationObjectTypeOfDE = (ConfigurationObjectType)model.getType("typ.de");

      List<SystemObject> list = __ConfigurationObjectType.getObjects();
      model.getConfigurationData(list, __AtgEAK);
    }
  }

}
