/*
 * Copyright (c) 2010 by inovat, innovative systeme - verkehr - tunnel - technik, Dipl.-Ing. H. C. Kniss
 *
 * This file is part of de.inovat.kex.bast.exporter.FTPServerConnection
 *
 * de.inovat.kex.bast.exporter.FTPServerConnection is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * de.inovat.kex.bast.exporter.FTPServerConnection is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * <p>
 * You should have received a copy of the GNU General Public License
 * along with de.inovat.kex.bast.exporter.FTPServerConnection; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 * Contact Information:
 * inovat, Dipl.-Ing. H. C. Kniss
 * Koelner Strasse 30
 * D-50859 Kln
 * +49 (0)2234 4301 800
 * info@invat.de
 * www.inovat.de
 */



package de.inovat.kex.bast.exporter;

//~ NICHT JDK IMPORTE =========================================================

import de.inovat.kex.bast.KonstantenBast.BAST_FORMAT;

//~ KLASSEN ===================================================================

/**
 * Die Klasse verwaltet die Informationen zur FTP-ServerConnection.
 * Die Klasse ist als Singelton realisiert, so dass die Serverinformationen global abgerufen werden knnen.
 * <p/>
 *
 * @author inovat, innovative systeme - verkehr - tunnel - technik
 * @author Dagmar Rostek (DR)
 * @version $Revision: 439 $ / $Date: 2010-09-29 08:13:47 +0200 (Mi, 29 Sep 2010) $ / ($Author: DR $)
 */
public class FTPServerConnection {

    /**
     * Instanz.
     */
    private static FTPServerConnection instance = null;

    //~ FELDER ================================================================

    /**
     * Boolesches Attribut das angibt, ob es sich um einen BASt-Band Export Achslast handelt.
     */

    // private boolean _achslast = false;

    /**
     * Das Attribut gibt an, wieviele Zeilen schon exportiert worden sind.
     */
    private int _exportiertAnzahl = 0;

    /**
     * Das Attribut beinhaltet den FTP-Benutzernamen.
     */
    private String _ftpBenutzerNameExport = "";

    /**
     * Das Attribut beinhaltet das Passwort fr den FTP-Server Zugriff.
     */
    private String _ftpPasswdExport = "";

    /**
     * Das Attribut gibt den FTP-Servernamen an. (host).
     */
    private String _ftpServerNameExport = "";

    /**
     * Das Attribut beinhaltet die Information, ob die Datei auf dem FTP-Server berschrieben
     * werden soll oder nicht, true wenn sie berschrieben werden soll, false sonst.
     */
    private boolean _dateiErsetzen = false;

    /**
     * Das Attribut beinhaltet die Information, ob es innerhalb eines Jahresverlaufes zu
     * nderungen der Datenstruktur gekommen ist, so dass der Dateiname anders zusammen
     * gesetzt werden muss, um diese nderung kenntlich zu machen.
     */
    private boolean _aenderungDatenstruktur = false;

    /**
     * Attribut Initiale Prozentwert.
     */
    private int _prozentExport = 2;

    /**
     * Boolesches Attribut das angibt, ob es sich um
     * einen Achslastdatenexport handelt.
     */
    private boolean _isAchslastExport = false;

    /** Anzahl der Langzeitzhlstellen die Exportiert werden. */
    private int _anzahlLZZ;

    /** Anzahl der Monate fr die ein Export durchgefhrt wird. Wird zur Berechnung der Progressbar bentigt. */
    private int _anzahlMonate;

    /**
     * Das Attribut beinhaltet die Information um welches BastBandExport Format es sich handelt.
     */
    private BAST_FORMAT _bastFormatExport;

    /** Legt fest, ob die FTP-Verbindung im PassivModus eingebaut werden soll. */
    private boolean _ftpPassivModus;

    /**
     * Anzahl der zu exportierenden Dateien.
     */
    private int _zuExportierenAnzahl;

    //~ METHODEN ==============================================================

    /**
     *
     *
     * Die Methode bestimmt den Prozentwert bzgl. der schon exportierten Dateien.
     * @return   Prozentwert als String.
     * Sollte der Prozentwert > 100 sein, wird 99 retuniert.
     * @param allLinesEinesExportes die Anzahl der Zeilen fr den aktuellen Export.
     */
    public String bestimmeProzentExport(int allLinesEinesExportes) {
        double value;

        _exportiertAnzahl++;

        if (_isAchslastExport) {

            // Verwendung eines Durchschnittwertes von 30 fr die Anzahl der Tage eines Monats.
            _zuExportierenAnzahl = ((_anzahlLZZ) * _anzahlMonate) * 30;
        } else {
            _zuExportierenAnzahl = ((_anzahlLZZ) * _anzahlMonate) * allLinesEinesExportes;
        }

        double prozentSchritt = (double) 100 / (double) _zuExportierenAnzahl;

        value = prozentSchritt * _exportiertAnzahl;

        if (!(_prozentExport > (2 + (int) value))) {
            _prozentExport = 2 + (int) value;
        }

        int ergebnis = _prozentExport;

        if ((ergebnis < 100) && (ergebnis >= 2)) {
            return Integer.toString(ergebnis);
        } else if (ergebnis < 2) {
            return "2";
        } else {
            return "99";
        }
    }

    /**
     * Die Methode initialisiert alle Parameter fr den BASt-Band Export bzgl. der FTP-Server angaben.
     *
     * @param ftpServerName          FTP-Servername.
     * @param ftpBenutzerName        FTP-Server Benutzername.
     * @param ftpPasswd              FTP-Server Passwort.
     * @param ftpPassivModus     Legt fest, ob die FTP-Verbindung im PassivModus eingebaut werden soll.
     * @param bastFormat             BASt-Band Format.
     * @param dateiErsetzen          Flag das angibt ob die Datei ersetzt werden soll.
     * @param aenderungDatenstruktur Flag das angibt, ob sich die Datenstruktur gendert hat.
     */
    public void FTPServerConnectionSetParameterExport(String ftpServerName, String ftpBenutzerName, String ftpPasswd, boolean ftpPassivModus, BAST_FORMAT bastFormat, boolean dateiErsetzen, boolean aenderungDatenstruktur) {
        _ftpServerNameExport    = ftpServerName;
        _ftpBenutzerNameExport  = ftpBenutzerName;
        _ftpPasswdExport        = ftpPasswd;
        _ftpPassivModus         = ftpPassivModus;
        _bastFormatExport       = bastFormat;
        _dateiErsetzen          = dateiErsetzen;
        _aenderungDatenstruktur = aenderungDatenstruktur;
    }

    //~ GET METHODEN ==========================================================

    /**
     * Getter Benutzername.
     *
     * @return Benutzername als String.
     */
    public String getFtpBenutzerNameExport() {
        return _ftpBenutzerNameExport;
    }

    /**
     * Getter fr das FTP-Passwort.
     *
     * @return Passwort als String.
     */
    public String getFtpPasswdExport() {
        return _ftpPasswdExport;
    }

    /**
     * Die Methode retuniert je nach BASt-Band Format den
     * Pfad auf den FTP-Server, indem die erzeugten Dateien exportiert werden sollen.
     *
     * @return String, der die Ausgabeverzeichnisstruktur enthlt.
     */
    public String getFtpPathExport() {
        try {
            _isAchslastExport = (_bastFormatExport == BAST_FORMAT.AXL_EF1);

            return _bastFormatExport._ftpExportOrdner;
        }
        catch (Exception e) {
            return "";
        }
    }

    /**
     * Die Methode retuniert den FTP-Servernamen als String.
     *
     * @return FTP Servernamen.
     */
    public String getFtpServerNameExport() {
        return _ftpServerNameExport;
    }

    /**
     * Retuniert Instanz FTPServerConnection.
     * @return  Instanz FTPServerConnection.
     */
    public static FTPServerConnection getInstance() {
        if (instance == null) {
            instance = new FTPServerConnection();
        }

        return instance;
    }

    //~ METHODEN ==============================================================

    /**
     * Die Methode initialisiert alle Progressbarabhngigen Variablen.
     */
    public void initialisiereProgressbarKonstanten() {
        _exportiertAnzahl    = 0;
        _zuExportierenAnzahl = 0;
        _prozentExport       = 2;
    }

    //~ GET METHODEN ==========================================================

    public boolean isAchslastExport() {
        return _isAchslastExport;
    }

    /**
     * Die Methode retuniert true, wenn eine nderung der Datenstruktur statt gefunden hat, false sonst.
     * @return true or false.
     */
    public boolean isAenderungDatenstruktur() {
        return _aenderungDatenstruktur;
    }

    /**
     *
     * @return  true wenn die Datei auf dem FTP-Server ersetzt werden soll, false sonst.
     */
    public boolean isDateiErsetzen() {
        return _dateiErsetzen;
    }

    public boolean isFtpPassivModus() {
        return _ftpPassivModus;
    }

    //~ SET METHODEN ==========================================================

    public void setAchslastExport(boolean achslastExport) {
        _isAchslastExport = achslastExport;
    }

    /**
     *  Setter Anzahl Langzeitzhlstellen.
     * @param anzahlLZZ  Anzahl der Langzeitzhlstellen.
     */
    public void setAnzahlLZZ(int anzahlLZZ) {
        _anzahlLZZ = anzahlLZZ;
    }

    /**
     * Setter Anzahl Monate.
     * @param anzahlMonate  anzahlMonate.
     */
    public void setAnzahlMonate(int anzahlMonate) {
        this._anzahlMonate = anzahlMonate + 1;
    }

    /**
     *
     * Setter der Anzahl der exportierten Zeilen beim Export.
     * Dieses Attribut muss vor jedem Neustart wieder auf 0 gesetzt werden.
     * @param exportiertAnzahl Anzahl der schon exportierten Zeilen.
     */
    public void setExportiertAnzahl(int exportiertAnzahl) {
        _exportiertAnzahl = exportiertAnzahl;
    }
}


//~Formatiert mit 'inovat Kodierkonvention' am 02.08.12
